@@ -21,13 +21,27 @@
  * SOFTWARE.
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <wordexp.h>
 #include <ctype.h>
 #include <unistd.h>
+
+#ifndef HAVE_ASPRINTF
+#include <compat/asprintf.h>
+#endif
+
+#ifdef HAVE_IOCTL
 #include <sys/ioctl.h>
+#endif
+
+#ifdef HAVE_WORDEXP
+#include <wordexp.h>
+#endif
 
 #include <lualib.h>
 #include <lauxlib.h>
@@ -53,20 +67,24 @@ extern char *readline ();
 /* This is a simple readline-like function in case readline is not
  * available. */
 
+#define MAXINPUT 1024
+
 static char *readline(char *prompt)
 {
     char *line = NULL;
-    size_t n;
     int k;
+
+    line = malloc (MAXINPUT);
     
     fputs(prompt, stdout);
     fflush(stdout);
-    k = getline (&line, &n, stdin);
-
-    if (k < 0) {
+    
+    if (!fgets(line, MAXINPUT, stdin)) {
 	return NULL;
     }
     
+    k = strlen (line);
+
     if (line[k - 1] == '\n') {
 	line[k - 1] = '\0';
     }
@@ -86,10 +104,10 @@ extern void add_history ();
 extern int write_history ();
 extern int read_history ();
 #  endif /* defined(HAVE_READLINE_HISTORY_H) */
-static char *logfile;
-
 #endif /* HAVE_READLINE_HISTORY */
 
+static char *logfile;
+
 #define print_output(...) fprintf (stdout, __VA_ARGS__), fflush(stdout)
 #define print_error(...) fprintf (stderr, __VA_ARGS__), fflush(stderr)
 #define absolute(_L, _i) (_i < 0 ? lua_gettop (_L) + _i + 1 : _i)
@@ -121,13 +139,11 @@ static char *chunkname, *prompts[2], *buffer = NULL;
 
 static void display_matches (char **matches, int num_matches, int max_length)
 {
-    /* print_output ("%s", COLOR(6)); */
+    print_output ("%s", COLOR(7));
     rl_display_match_list (matches, num_matches, max_length);
-    putc ('\n', stdout);
-    /* print_output ("%s", COLOR(0)); */
-    /* rl_redisplay(); */
+    print_output ("%s", COLOR(0));
+    rl_on_new_line ();
 }
-
     
 static char *keyword_completions (const char *text, int state)
 {
@@ -207,17 +223,17 @@ static char *table_key_completions (const char *text, int state)
 	    return NULL;
 	}
     }
-    
+   
     /* Iterate the table/userdata and generate matches. */
     
     while (lua_pushvalue(_L, -3), lua_insert (_L, -3),
 	   lua_pushvalue(_L, -2), lua_insert (_L, -4),
-	   lua_pcall (_L, 2, LUA_MULTRET, 0) == 0) {
+	   lua_pcall (_L, 2, 2, 0) == 0) {
 	char *candidate;
 	size_t l, m;
 	int nospace;
 
-	if (lua_isnil(_L, -1)) {
+	if (lua_isnil(_L, -2)) {
 	    return NULL;
 	}
 
@@ -296,7 +312,7 @@ static char *table_key_completions (const char *text, int state)
 static char *generator (const char *text, int state)
 {
     static int which;
-    char *match;
+    char *match = NULL;
 
     if (state == 0) {
 	which = 0;
@@ -317,6 +333,45 @@ static char *generator (const char *text, int state)
     
     if (which == 1) {
 	match = table_key_completions (text, state);
+
+	if (!match) {
+	    which = 2;
+	    state = 0;
+	}
+    }
+
+    /* Try to complete a filename. */
+    
+    if (which == 2) {
+	const char *start;
+
+	/* Ignore any unquoted characters, we'll only be trying to
+	 * complete file names inside strings. */
+	
+	for (start = text;
+	     *start && *(start - 1) != '\'' && *(start - 1) != '"';
+	     start += 1);
+
+	if (*start != '\0') {
+	    int n;
+
+	    /* Don't append a space at the end of the match.  It isn't
+	     * very helpful in this context. */
+	    
+	    rl_completion_suppress_append = 1;
+	    
+	    match = rl_filename_completion_function (start, state);
+	    if (match) {
+		/* If a match was produced, prepend the unquoted
+		 * characters. */
+		
+		n = strlen (match) + 1;
+	    
+		match = realloc (match, n + start - text);
+		memmove (match + (start - text), match, n);
+		strncpy (match, text, start - text);
+	    }
+	}
     }
 
     return match;
@@ -421,194 +476,6 @@ static int execute ()
     return status;
 }
 
-int luap_call (lua_State *L, int n) {
-    int h, status;
-
-    /* Push the error handler onto the stack. */
-    
-    h = lua_gettop(L) - n;
-    lua_pushcfunction (L, traceback);
-    lua_insert (L, h);
-    
-    /* Try to execute the supplied chunk and keep note of any return
-     * values. */
-
-    status = lua_pcall(L, n, LUA_MULTRET, h);
-
-    /* Print all returned values with proper formatting. */
-    
-    if (status != LUA_OK) {
-	print_error ("%s%s%s\n", COLOR(1), lua_tostring (L, -1), COLOR(0));
-	lua_pop (L, 1);
-    }
-
-    lua_remove (L, h);
-
-    return status;
-}
-
-void luap_setprompts(lua_State *L, const char *single, char *multi)
-{
-    prompts[0] = realloc (prompts[0], strlen (single) + 16);
-    sprintf (prompts[0], "\001%s\002%s\001%s\002", COLOR(6), single, COLOR(0));
-
-    prompts[1] = realloc (prompts[1], strlen (multi) + 16);
-    sprintf (prompts[1], "\001%s\002%s\001%s\002", COLOR(6), multi, COLOR(0));
-}
-
-void luap_sethistory(lua_State *L, const char *file)
-{
-    wordexp_t words;
-
-    wordexp(file, &words, 0);
-
-    if (words.we_wordc != 1) {
-	logfile = NULL;
-    } else {
-	logfile = realloc (logfile, strlen(words.we_wordv[0]) + 1);
-	strcpy (logfile, words.we_wordv[0]);
-    }
-
-    wordfree (&words);
-}
-
-void luap_setcolor(lua_State *L, int enable)
-{
-    colorize = enable;
-}
-
-void luap_setname(lua_State *L, const char *name)
-{
-    chunkname = realloc (chunkname, strlen(name) + 2);
-    chunkname[0] = '=';
-    strcpy (chunkname + 1, name);
-}
-
-void luap_enter(lua_State *L)
-{
-    int incomplete = 0, s = 0, t = 0, l;
-    char *line, *new;
-
-    /* Save the state since it needs to be passed to some readline
-     * callbacks. */
-    
-    _L = L;
-    
-    if (!initialized) {
-#ifdef HAVE_LIBREADLINE
-	rl_basic_word_break_characters = " \t\n`@$><=;|&{(";
-	rl_attempted_completion_function = complete;
-	rl_completion_display_matches_hook = display_matches;
-#endif
-	
-#ifdef HAVE_READLINE_HISTORY
-	/* Load the command history if there is one. */
-    
-	if (logfile) {
-	    read_history (logfile);
-	}
-#endif
-	if (!chunkname) {
-	    luap_setname (L, "lua");
-	}
-	
-	if (!prompts[0]) {
-	    luap_setprompts (L, ">  ", ">> ");
-	}
-	
-	atexit (finish);
-
-	initialized = 1;
-    }
-
-    while ((line = readline (incomplete ? prompts[1] : prompts[0]))) {
-	int status;
-	
-	if (*line == '\0') {
-	    continue;
-	}
-
-#ifdef HAVE_READLINE_HISTORY
-	/* Add the line to the history if non-empty. */
-	    add_history (line);
-#endif
-
-	/* Add/copy the line to the buffer. */
-	    
-	if (incomplete) {
-	    s += strlen (line) + 1;
-
-	    if (s > t) {
-		buffer = realloc (buffer, s + 1);
-		t = s;
-	    }
-		    
-	    strcat (buffer, "\n");
-	    strcat (buffer, line);
-	} else {
-	    s = strlen (line);
-
-	    if (s > t) {
-		buffer = realloc (buffer, s + 1);
-		t = s;
-	    }
-		    
-	    strcpy (buffer, line);
-	}
-
-	/* Try to execute the line with a return prepended first.  If
-	 * this works we can show returned values. */
-
-	l = asprintf (&new, "return %s", buffer);
-
-	if (luaL_loadbuffer(_L, new, l, chunkname) == LUA_OK) {
-	    execute();
-	} else {
-	    lua_pop (_L, 1);
-
-	    /* Try to execute the line as-is. */
-
-	    status = luaL_loadbuffer(_L, buffer, s, chunkname);
-		
-	    incomplete = 0;
-
-	    if (status == LUA_ERRSYNTAX) {
-		const char *message;
-		const int k = sizeof(EOF_MARKER) / sizeof(char) - 1;
-		size_t n;
-
-		message = lua_tolstring (_L, -1, &n);
-
-		/* If the error message mentions an unexpected eof
-		 * then consider this a multi-line statement and wait
-		 * for more input.  If not then just print the error
-		 * message.*/
-		
-		if (n > k &&
-		    !strncmp (message + n - k, EOF_MARKER, k)) {
-		    incomplete = 1;
-		} else {
-		    print_error ("%s%s%s\n", COLOR(1), lua_tostring (_L, -1), COLOR(0));
-		}
-
-		lua_pop (_L, 1);
-	    } else if (status == LUA_ERRMEM) {
-		print_error ("%s%s%s\n", COLOR(1), lua_tostring (_L, -1), COLOR(0));
-		lua_pop (_L, 1);
-	    } else {
-		/* Try to execute the loaded chunk. */
-		
-		execute ();
-		incomplete = 0;
-	    }
-	}
-	
-	free (line);
-    }
-
-    print_output ("\n");
-}
-
 /* This is the pretty-printing related stuff. */
 
 static char *dump;
@@ -766,8 +633,8 @@ static void describe (lua_State *L, int index)
 		if (s[i] == '=') {
 		    level += 1;
 		} else if (s[i] == ']') {
-		    if (level == uselevel) {
-			uselevel += 1;
+		    if (level >= uselevel) {
+			uselevel = level + 1;
 		    }
 		} else {
 		    started = 0;
@@ -858,7 +725,7 @@ static void describe (lua_State *L, int index)
 		      COLOR(7), COLOR(8), lua_topointer (L, index));
 	dump_string (s, n);
     } else if (type == LUA_TTABLE) {
-	int i, l, n, oldindent, multiline;
+	int i, l, n, oldindent, multiline, nobreak;
 
 	/* Check if table is too deeply nested. */
 	
@@ -899,10 +766,10 @@ static void describe (lua_State *L, int index)
 	    lua_pop (L, 1);
 	}
 
-	/* Add the table to the ancestor list and pop it (the ancestor
-	 * list table). */
+	/* Add the table to the ancestor list and pop the ancestor
+	 * list table. */
 	
-	lua_pushvalue (L, -2);
+	lua_pushvalue (L, index);
 	lua_rawseti (L, -2, n + 1);
 	lua_pop (L, 1);
 	
@@ -913,6 +780,7 @@ static void describe (lua_State *L, int index)
 	oldindent = indent;
 	indent = column;
 	multiline = 0;
+	nobreak = 0;
 	
 	l = lua_rawlen (L, index);
 
@@ -922,14 +790,19 @@ static void describe (lua_State *L, int index)
 	    lua_pushinteger (L, i + 1);
 	    lua_gettable (L, index);
 
-	    /* Start a fresh lines when dumping tables to make sure
+	    /* Start a fresh line when dumping tables to make sure
 	     * there's plenty of room. */
 	    
 	    if (lua_istable (L, -1)) {
-		break_line();
+		if (!nobreak) {
+		    break_line();
+		} 
+
 		multiline = 1;
 	    }
 
+	    nobreak = 0;
+
 	    /* Dump the value and separating comma. */
 	    
 	    describe (L, -1);
@@ -937,6 +810,7 @@ static void describe (lua_State *L, int index)
 
 	    if (lua_istable (L, -1) && i != l - 1) {
 		break_line();
+		nobreak = 1;
 	    }
 		
 	    lua_pop (L, 1);
@@ -946,7 +820,7 @@ static void describe (lua_State *L, int index)
 	
 	lua_pushnil (L);
 	while (lua_next (L, index) != 0) {
-	    if (!lua_isnumber (L, -2) ||
+	    if (!lua_type (L, -2) == LUA_TNUMBER ||
 	    	lua_tonumber (L, -2) != lua_tointeger (L, -2) ||
 		lua_tointeger (L, -2) < 1 ||
 		lua_tointeger (L, -2) > l) {
@@ -987,6 +861,13 @@ static void describe (lua_State *L, int index)
 	    lua_pop (L, 1);
 	}
 
+	/* Remove the table from the ancestor list. */
+	
+	lua_rawgeti (L, LUA_REGISTRYINDEX, ancestors);
+	lua_pushnil (L);
+	lua_rawseti (L, -2, n + 1);
+	lua_pop (L, 1);
+	
 	/* Pop the indentation level. */
 	
 	indent = oldindent;
@@ -1002,10 +883,9 @@ static void describe (lua_State *L, int index)
 
 char *luap_describe (lua_State *L, int index)
 {
+#ifdef HAVE_IOCTL
     struct winsize w;
 
-    index = absolute (L, index);    
-
     /* Initialize the state. */
     
     if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &w) < 0) {
@@ -1013,7 +893,11 @@ char *luap_describe (lua_State *L, int index)
     } else {
 	linewidth = w.ws_col;
     }
-
+#else
+    linewidth = 80;
+#endif
+    
+    index = absolute (L, index);
     offset = 0;
     indent = 0;
     column = 0;
@@ -1030,3 +914,245 @@ char *luap_describe (lua_State *L, int index)
     
     return dump;
 }
+
+/* These are custom commands. */
+
+#ifdef HAVE_LIBREADLINE
+static int describe_stack (int count, int key)
+{
+    int i, h;
+    
+    print_output ("%s", COLOR(7));
+
+    h = lua_gettop (_L);
+
+    if (count < 0) {
+	i = h + count + 1;
+	
+	if (i > 0 && i <= h) {
+	    print_output ("\nValue at stack index %d(%d):\n%s%s",
+			  i, -h + i - 1, COLOR(3), luap_describe (_L, i));
+	} else {
+	    print_error ("Invalid stack index.\n");
+	}
+    } else {
+	if (h > 0) {
+	    print_output ("\nThe stack contains %d values.\n", h);
+	    for (i = 1 ; i <= h ; i += 1) {
+		print_output ("\n%d(%d):\t%s", i, -h + i - 1, lua_typename(_L, lua_type(_L, i)));
+	    }
+	} else {
+	    print_output ("\nThe stack is empty.");
+	}
+    }
+    
+    print_output ("%s\n", COLOR(0));
+
+    rl_on_new_line ();
+
+    return 0;
+}
+#endif
+
+int luap_call (lua_State *L, int n) {
+    int h, status;
+
+    /* Push the error handler onto the stack. */
+    
+    h = lua_gettop(L) - n;
+    lua_pushcfunction (L, traceback);
+    lua_insert (L, h);
+    
+    /* Try to execute the supplied chunk and keep note of any return
+     * values. */
+
+    status = lua_pcall(L, n, LUA_MULTRET, h);
+
+    /* Print all returned values with proper formatting. */
+    
+    if (status != LUA_OK) {
+	print_error ("%s%s%s\n", COLOR(1), lua_tostring (L, -1), COLOR(0));
+	lua_pop (L, 1);
+    }
+
+    lua_remove (L, h);
+
+    return status;
+}
+
+void luap_setprompts(lua_State *L, const char *single, char *multi)
+{
+    prompts[0] = realloc (prompts[0], strlen (single) + 16);
+    prompts[1] = realloc (prompts[1], strlen (multi) + 16);
+#ifdef HAVE_LIBREADLINE
+    sprintf (prompts[0], "\001%s\002%s\001%s\002",
+	     COLOR(6), single, COLOR(0));
+    sprintf (prompts[1], "\001%s\002%s\001%s\002",
+	     COLOR(6), multi, COLOR(0));
+#else
+    sprintf (prompts[0], "%s%s%s", COLOR(6), single, COLOR(0));
+    sprintf (prompts[0], "%s%s%s", COLOR(6), multi, COLOR(0));
+#endif
+}
+
+void luap_sethistory(lua_State *L, const char *file)
+{
+#ifdef HAVE_WORDEXP
+    wordexp_t words;
+
+    wordexp(file, &words, 0);
+
+    if (words.we_wordc != 1) {
+	logfile = NULL;
+    } else {
+	logfile = realloc (logfile, strlen(words.we_wordv[0]) + 1);
+	strcpy (logfile, words.we_wordv[0]);
+    }
+
+    wordfree (&words);
+#else
+    logfile = realloc (logfile, strlen(file) + 1);
+    strcpy (logfile, file);
+#endif
+}
+
+void luap_setcolor(lua_State *L, int enable)
+{
+    colorize = enable;
+}
+
+void luap_setname(lua_State *L, const char *name)
+{
+    chunkname = realloc (chunkname, strlen(name) + 2);
+    chunkname[0] = '=';
+    strcpy (chunkname + 1, name);
+}
+
+void luap_enter(lua_State *L)
+{
+    int incomplete = 0, s = 0, t = 0, l;
+    char *line, *new;
+
+    /* Save the state since it needs to be passed to some readline
+     * callbacks. */
+    
+    _L = L;
+    
+    if (!initialized) {
+#ifdef HAVE_LIBREADLINE
+	rl_basic_word_break_characters = " \t\n`@$><=;|&{(";
+	rl_attempted_completion_function = complete;
+	rl_completion_display_matches_hook = display_matches;
+
+	rl_add_defun ("lua-describe-stack", describe_stack, META('s'));
+#endif
+	
+#ifdef HAVE_READLINE_HISTORY
+	/* Load the command history if there is one. */
+    
+	if (logfile) {
+	    read_history (logfile);
+	}
+#endif
+	if (!chunkname) {
+	    luap_setname (L, "lua");
+	}
+	
+	if (!prompts[0]) {
+	    luap_setprompts (L, ">  ", ">> ");
+	}
+	
+	atexit (finish);
+
+	initialized = 1;
+    }
+
+    while ((line = readline (incomplete ? prompts[1] : prompts[0]))) {
+	int status;
+	
+	if (*line == '\0') {
+	    continue;
+	}
+
+#ifdef HAVE_READLINE_HISTORY
+	/* Add the line to the history if non-empty. */
+	    add_history (line);
+#endif
+
+	/* Add/copy the line to the buffer. */
+	    
+	if (incomplete) {
+	    s += strlen (line) + 1;
+
+	    if (s > t) {
+		buffer = realloc (buffer, s + 1);
+		t = s;
+	    }
+		    
+	    strcat (buffer, "\n");
+	    strcat (buffer, line);
+	} else {
+	    s = strlen (line);
+
+	    if (s > t) {
+		buffer = realloc (buffer, s + 1);
+		t = s;
+	    }
+		    
+	    strcpy (buffer, line);
+	}
+
+	/* Try to execute the line with a return prepended first.  If
+	 * this works we can show returned values. */
+
+	l = asprintf (&new, "return %s", buffer);
+
+	if (luaL_loadbuffer(_L, new, l, chunkname) == LUA_OK) {
+	    execute();
+		
+	    incomplete = 0;
+	} else {
+	    lua_pop (_L, 1);
+
+	    /* Try to execute the line as-is. */
+
+	    status = luaL_loadbuffer(_L, buffer, s, chunkname);
+		
+	    incomplete = 0;
+
+	    if (status == LUA_ERRSYNTAX) {
+		const char *message;
+		const int k = sizeof(EOF_MARKER) / sizeof(char) - 1;
+		size_t n;
+
+		message = lua_tolstring (_L, -1, &n);
+
+		/* If the error message mentions an unexpected eof
+		 * then consider this a multi-line statement and wait
+		 * for more input.  If not then just print the error
+		 * message.*/
+		
+		if (n > k &&
+		    !strncmp (message + n - k, EOF_MARKER, k)) {
+		    incomplete = 1;
+		} else {
+		    print_error ("%s%s%s\n", COLOR(1), lua_tostring (_L, -1), COLOR(0));
+		}
+
+		lua_pop (_L, 1);
+	    } else if (status == LUA_ERRMEM) {
+		print_error ("%s%s%s\n", COLOR(1), lua_tostring (_L, -1), COLOR(0));
+		lua_pop (_L, 1);
+	    } else {
+		/* Try to execute the loaded chunk. */
+		
+		execute ();
+		incomplete = 0;
+	    }
+	}
+	
+	free (line);
+    }
+
+    print_output ("\n");
+}
